import datetime
import json
import dns.name
import postgresql
import ipaddress
import requests

class BooleanSearchFilter(object):
    def __init__(self, name, default, form_name):
        self.name = name
        self.default = default
        self.form_name = form_name


# Database/general objects

class DBObject(object):
    TABLE = None

    def get_event_log(self, db, connection, bnd_entity_id, filter_login_name=None, top_n=5):
        if self.TABLE is None or self.TABLE.startswith('META'):
            raise NotImplementedError()
        stmt = """
            select
              fact.log_ts as timestamp,
              evdt.short_name as evlog_dflt_typ_short_name,
              evdt.name as evlog_dflt_typ_name, 
              evdt.confirmed_action as evlog_dflt_typ_conf_action,
              cm.login_name as cntl_mgr_login_name,
              cm.first_name as cntl_mgr_first_name,
              cm.last_name as cntl_mgr_last_name,
              cm.email as cntl_mgr_email,
              cp.name as cntl_pkg_name,
              ct.name as cntl_tab_name,
              ct.cont_sing1 as cntl_tab_sing_1_title,
              ct.cont_plur1 as cntl_tab_plur1_title,
              ct.cont_cmpnd as cntl_tab_cmpnd_title,
              fact.log_data as log_data
            from evlog.calc_top_n_log_ta(
                in_cntl_tab_name => {bnd_entity_name},
                in_evlog_dflt_typ_short_name => {bnd_log_typ},
                in_cntl_mgr_login_name => {bnd_log_mgr},
                in_oltp_pk => (select key_nr from dns.get_range_rec({bnd_entity_id})),
                in_top_n => {bnd_top_n}
              ) AS fact
              JOIN evlog_row2dwh dim_row ON fact.dwh_row_key_nr = dim_row.dwh_key_nr
              JOIN evlog_dflt_typ2dwh dim_typ ON fact.dwh_typ_key_nr = dim_typ.dwh_key_nr
              JOIN cntl_mgr2dwh dim_mgr ON fact.dwh_mgr_key_nr = dim_mgr.dwh_key_nr
              JOIN cntl_tab ct ON ct.key_nr = dim_row.oltp_tab_key_nr
              JOIN cntl_pkg cp ON cp.key_nr = ct.cntl_pkg_key_nr
              JOIN evlog_dflt_typ evdt ON evdt.key_nr = dim_typ.oltp_key_nr
              LEFT JOIN cntl_mgr cm ON cm.key_nr = dim_mgr.oltp_key_nr
            where fact.evlog_dflt_key_nr is null
            order by fact.log_ts desc, ct.name
        """
        bindings = {'bnd_log_mgr': None, 'bnd_log_typ': filter_login_name, 'bnd_top_n': top_n,
                    'bnd_entity_name': self.TABLE,
                    'bnd_entity_id': bnd_entity_id}
        res = db.execute(connection, stmt, bindings)
        return [EVLogEntry(mgr=DBMgr(**_marshal_dict(r, DBMgr)), **r) for r in res]


class MetaDBObject(DBObject):
    TABLE = "META_"


def _marshal_dict_str(dict, string: str):
    return {k.replace("{}_".format(string), ""): v for k, v in dict.items()}


def _marshal_dict(dict, obj: DBObject):
    return _marshal_dict_str(dict, obj.TABLE.lower())


class RRType(MetaDBObject):
    def __init__(self, name, is_intern):
        self.name = name
        self.is_intern = is_intern

    @staticmethod
    def get_all(db, connection):
        rr_types = db.execute(connection, "SELECT name, is_intern FROM dns_rr_type order by name", {})
        return [RRType(**r) for r in rr_types]


class EVLogEntry(DBObject):
    TABLE = 'v_evlog_dflt_top_n'

    def __init__(self, timestamp=None, evlog_dflt_typ_short_name=None, evlog_dflt_typ_name=None,
                 evlog_dflt_typ_conf_action=None, mgr=None, cntl_pkg_name=None, cntl_tab_name=None,
                 cntl_tab_sing1_title=None, cntl_tab_plur1_title=None, cntl_tab_cmpnd_title=None, log_data=None,
                 **args):
        self.timestamp = timestamp
        self.evlog_dflt_typ_short_name = evlog_dflt_typ_short_name
        self.evlog_dflt_typ_name = evlog_dflt_typ_name
        self.evlog_dflt_typ_conf_action = evlog_dflt_typ_conf_action
        self.cntl_pkg_name = cntl_pkg_name
        self.cntl_tab_name = cntl_tab_name
        self.cntl_tab_sing1_title = cntl_tab_sing1_title
        self.cntl_tab_plur1_title = cntl_tab_plur1_title
        self.cntl_tab_cmpnd_title = cntl_tab_cmpnd_title
        self.log_data = log_data
        self.mgr = mgr


class Mgr(object):
    is_scc_itb = False

    def __init__(self, login_name, first_name, last_name, email, **args):
        self.login_name = login_name
        self.first_name = first_name
        self.last_name = last_name
        self.email = email

    def get_areas(self, **kwargs):
        raise NotImplementedError()

    def update_acd_cache(self, **kwargs):
        raise NotImplementedError()

    def get_possible_rr_inttypes(self, **kwargs):
        raise NotImplementedError()

    def __eq__(self, other):
        return type(self) == type(other) and self.login_name == other.login_name


class SimpleKITUser(Mgr):
    is_scc_itb = False

    def __init__(self, login_name, first_name, last_name, email, **args):
        self.login_name = login_name
        self.first_name = first_name
        self.last_name = last_name
        self.email = email

    def get_areas(self, **kwargs):
        return []

    def get_possible_rr_inttypes(self, **kwargs):
        return []

    def update_acd_cache(self, **kwargs):
        pass


class VLAN(MetaDBObject):
    TABLE = 'META_vlan'

    def __init__(self, vlan_id, vlan_name, bcd_name, ni_name, **args):
        self.id = vlan_id
        self.name = vlan_name
        self.bcd_name = bcd_name
        self.ni_name = ni_name


class DBMgr(Mgr, DBObject):
    TABLE = 'cntl_mgr'

    def __init__(self, login_name, first_name, last_name, email, adm_lock=None, kit_status=None, is_scc_itb=None,
                 acd_cache=None, key_nr=None,
                 **args):
        super().__init__(login_name=login_name, first_name=first_name, last_name=last_name, email=email)
        self.key_nr = key_nr
        # kist_status -1: not yet defined
        self.is_scc_itb = is_scc_itb
        self.is_locked = kit_status == 0 or adm_lock
        self.__acd_cache = json.loads(acd_cache) if acd_cache is not None else None

    def update_acd_cache(self, db, connection):
        res = db.execute(connection, """
            select init_mgr_session as cache from cntl.init_mgr_session(in_login_name:={login_name}, in_force_perm_check:=False)
        """, {'login_name': self.login_name})
        js = json.loads(res[0]['cache'])
        self.__acd_cache = js['acd_cache']
        self.key_nr = js['key_nr']
        self.email = js['email']
        self.is_scc_itb = js['is_scc_itb']
        self.is_locked = js['adm_lock']

    @staticmethod
    def get_by_login_name(db, connection, login_name):
        res = db.execute(connection, "SELECT * FROM cntl_mgr WHERE login_name={login_name}", {'login_name': login_name})
        return DBMgr(**res[0]) if len(res) > 0 else None

    def get_areas(self, db, connection):
        res = db.execute(connection,
                         DBNetArea.FULL_SELECT +
                         "LEFT JOIN dns_mgr2range on dns_range.key_nr = dns_mgr2range.dns_range_key_nr "
                         "where dns_mgr2range.cntl_mgr_key_nr = {user_id} order by vw_oe_short_name asc, META_vlan_ni_name, META_vlan_vlan_id, dns_range.name, "
                         "dns_range.suffix, nd_ip_subnet.subnet_type",
                         {'user_id': self.key_nr})
        return [DBNetArea(**r) for r in res]

    def get_possible_fqdn_types(self, db, connection):
        res = db.execute(connection, """
            select
            vnt.name, vnt.description
            from v_dns_dbnt vnt
            JOIN cntl_mgr m ON (m.acd_cache #> '{{cntl, xup}}' ? vnt.nt_cntl_perm_name or vnt.nt_cntl_perm_name is null)
            where m.login_name = {login_name}
            order by vnt.name_sortby

        """, {'login_name': self.login_name})
        return [FqdnNameType(**r) for r in res]

    def get_possible_rr_inttypes(self, db, connection):
        res = db.execute(connection, "select "
                                     "vrrd.name, vrrd.description, vrrd.dns_rr_type_name, vrrd.rrd_auto_dml_fqdn_text "
                                     "from v_dns_dbrt vrrd "
                                     "INNER JOIN cntl_mgr m ON (m.acd_cache #> '{{cntl, xup}}' ? vrrd.rrd_cntl_perm_name or vrrd.rrd_cntl_perm_name is null) "
                                     "where m.login_name = {login_name} and not dns_rr_type_force_auto_dml "
                                     "order by dns_rr_type_name, l_nt_name_sortby, r_nt_name_sortby, r_at, l_unq desc, r_unq desc",
                         {'login_name': self.login_name})
        return [IntType(**r) for r in res]

    def __eq__(self, other):
        return type(self) == type(other) and self.login_name == other.login_name

    def __hash__(self):
        return "{}_{}".format(self.key_nr, self.login_name)

    def get_fqdn(self, db, connection, fqdn):
        res = db.execute(connection, """
               SELECT fqdn_has_rr as dns_ntree_fqdn_has_rr, (dbnt_rec).name as dns_ntree_dbnt_name, fqdn_sub_count as dns_ntree_fqdn_sub_count, parent_fqdn as dns_ntree_parent_fqdn,
                                fqdn as dns_ntree_fqdn, fqdn_description as dns_ntree_description, (dbnt_rec).is_d as dns_ntree_is_d
                                FROM dns.list_fqdn(in_login_name:={login_name}, in_value:={fqdn}, in_listopt_subtree:=false, in_listopt_global:={global})
               """, {'login_name': self.login_name, 'global': self.has_permission('dns.ro'), 'fqdn': fqdn})
        return NTree(**_marshal_dict(res[0], NTree)) if len(res) > 0 else None

    def get_fqdn_root_nodes(self, db, connection):
        res = db.execute(connection, """
              select
              n.is_d AS dns_ntree_is_d,
              (rro.key_nr is not null) AS dns_ntree_fqdn_has_rr,
              count(distinct n_sub.key_nr) AS dns_ntree_fqdn_sub_count,
              pn.fqdn AS dns_ntree_parent_fqdn,
              n.fqdn AS dns_ntree_fqdn,
              n.description AS dns_ntree_description
            from dns_ntree n
              JOIN dns_ntree pn ON pn.key_nr = n.parent_key_nr
              LEFT JOIN dns_ntree n_sub ON n_sub.parent_key_nr = n.key_nr
              LEFT JOIN dns_rr_owner rro ON rro.dns_ntree_key_nr = n.key_nr
            where n.key_nr = any( {bnd_domain_nspc_pks}::bigint[] )
            group by
              n.is_d,
              dns_ntree_fqdn_has_rr,
              pn.fqdn,
              n.fqdn,
              n.description
            order by dns_ntree_fqdn
        """, {'bnd_domain_nspc_pks': [int(k) for k in self.__acd_cache['dns']['m2d'].keys()]})
        return [NTree(**_marshal_dict(r, NTree)) for r in res]

    def get_fqdn_nodes(self, root, db, connection):
        res = db.execute(connection, """
        SELECT fqdn_has_rr as dns_ntree_fqdn_has_rr, fqdn_sub_count as dns_ntree_fqdn_sub_count, parent_fqdn as dns_ntree_parent_fqdn,
                         fqdn as dns_ntree_fqdn, fqdn_description as dns_ntree_description, (dbnt_rec).is_d as dns_ntree_is_d,
                         (dbnt_rec).name as dns_ntree_dbnt_name
                         FROM dns.list_fqdn(in_login_name:={login_name}, in_parent_value_list:={root}, in_listopt_subtree:=false, in_listopt_global:={global})
        """, {'login_name': self.login_name, 'global': self.has_permission('dns.ro'), 'root': [root]})
        return [NTree(**_marshal_dict(r, NTree)) for r in res]

    def has_range(self, range_key_nr):
        return str(range_key_nr) in self.__acd_cache['dns']['m2r']

    def has_permission(self, permission):
        return permission in self.__acd_cache['cntl']['xup']

    def create_session_token(self, db, connection, stay_logged_in=False):
        descr = 'Session-Token ({})'.format(datetime.datetime.now())
        res = db.execute(connection, """
SELECT out_token_text as token, out_token_pk as pk FROM cntl.create_sess_auth({login_name}, {descr}, {stay_logged_in})
        """, {'login_name': self.login_name, 'stay_logged_in': stay_logged_in,
              'descr': descr})
        res = res[0]
        return APIToken(**res, login_name=self.login_name, description=descr)


class IntType(MetaDBObject):
    def __init__(self, name, description, dns_rr_type_name, rrd_auto_dml_fqdn_text=None, **kvargs):
        self.name = name
        self.description = description
        self.dns_rr_type_name = dns_rr_type_name
        self.left = self.right = self.right_bits = None
        self.is_set = False
        self.rrd_auto_dml_fqdn_text = rrd_auto_dml_fqdn_text
        if name is not None:
            # TODO use dns_rr_def.r.unq instead of split
            split = name.split(':')
            self.left = split[0] + ':' + split[1]
            self.right = split[2]
            self.right_bits = self.right.split(',')[1]
            self.is_set = len(self.right_bits) > 0 and (self.right_bits[-1] == '0' or self.right_bits[-1] == '2')

    def __eq__(self, other):
        if type(other) is not IntType:
            return False
        if other.name is None and self.name is None:
            return other.dns_rr_type_name == self.dns_rr_type_name
        return other.name == self.name


class FqdnNameType(object):
    def __init__(self, name, description, **kvargs):
        self.name = name
        self.description = description

    @staticmethod
    def get_all(db, connection):
        res = db.execute(connection, """
            select
            vnt.name, vnt.description
            from v_dns_dbnt vnt
        """, {})
        return [FqdnNameType(**r) for r in res]


class NetArea(object):
    def __init__(self, name, suffix, **args):
        self.name = name if not name.endswith('/') else name[:-1]
        self.suffix = int(suffix)
        self.key_nr = None
        self.description = None
        self.facility = None
        self.net = None
        self.print_name = self.name + '/' + str(suffix)
        self.dns_ntree_key_nr = None

    def __str__(self):
        return self.print_name

    def get_mgrs(self):
        raise NotImplementedError()

    def get_n_trees(self):
        raise NotImplementedError()

    def get_rr(self):
        raise NotImplementedError()

    def get_gateways(self):
        raise NotImplementedError()

    def get_addrs(self):
        raise NotImplementedError()

    def remove_mgrs(self):
        raise NotImplementedError()

    def add_mgrs(self):
        raise NotImplementedError()


class DBNetArea(DBObject, NetArea):
    TABLE = 'dns_range'

    FULL_SELECT = """
    SELECT
        dns_range. *,
        vw_oe.name as vw_oe_name,
        vw_oe.short_name as vw_oe_short_name,
        nd_ip_subnet.net as nd_ip_subnet_net,
        nd_ip_subnet.descr as nd_ip_subnet_descr,
        nd_vlan.name as META_vlan_vlan_name,
        nd_vlan.id as META_vlan_vlan_id,
        nd_bcd.name as META_vlan_bcd_name,
        nd_net_instnc.name as META_vlan_ni_name
    FROM dns_range JOIN
        vw_oe on dns_range.vw_oe_key_nr = vw_oe.key_nr JOIN
        nd_ip_subnet on dns_range.nd_ip_subnet_key_nr = nd_ip_subnet.key_nr JOIN
        nd_bcd ON nd_ip_subnet.nd_bcd_key_nr = nd_bcd.key_nr LEFT JOIN
        nd_vlan ON nd_vlan.nd_bcd_key_nr = nd_bcd.key_nr LEFT JOIN
        nd_net_instnc ON nd_vlan.nd_net_instnc_key_nr = nd_net_instnc.key_nr
"""

    def __init__(self, key_nr, vw_oe_key_nr, nd_ip_subnet_key_nr, nd_ip_subnet_type,
                 nd_net_instnc_is_priv, type, name, suffix, description, **args):
        super().__init__(name=name, suffix=suffix)
        self.key_nr = key_nr
        self.description = description
        self.facility = Facility(**_marshal_dict(args, Facility))
        self.net = Net(**_marshal_dict(args, Net))
        self._nat_range = None
        if args.get('meta_vlan_vlan_id', None) is not None:
            self.vlan = VLAN(**_marshal_dict(args, VLAN))
        else:
            self.vlan = None

    def get_mgrs(self, db, connection):
        res = db.execute(connection,
                         "SELECT cntl_mgr.* FROM cntl_mgr INNER JOIN dns_mgr2range ON cntl_mgr.key_nr = cntl_mgr_key_nr "
                         "WHERE dns_mgr2range.dns_range_key_nr = {dns_range_key_nr} ORDER BY kit_status DESC, last_name ASC",
                         {'dns_range_key_nr': self.key_nr})
        return [DBMgr(**r) for r in res]

    @staticmethod
    def get_by_name(db, connection, print_name=None, name=None, suffix=None):
        try:
            if print_name is not None:
                s = print_name.split('/')
                name = s[0]
                suffix = int(s[1])
            res = db.execute(connection,
                             DBNetArea.FULL_SELECT + " where dns_range.name={name} AND dns_range.suffix={suffix}",
                             {'name': name + '/', 'suffix': int(suffix)})
            if len(res) == 0:
                return None
        except ValueError:
            return None
        return DBNetArea(**res[0])

    @staticmethod
    def get_by_key_nr(db, connection, key_nr):
        res = db.execute(connection,
                         DBNetArea.FULL_SELECT + " where dns_range.key_nr={key_nr}",
                         {'key_nr': key_nr})
        if len(res) == 0:
            return None
        return DBNetArea(**res[0])

    def get_next_free_address_after(self, db, connection, addr):
        res = db.execute(connection,
                         """
                         SELECT ip_addr from dns_addr 
                         left join dns_rr_dst on dns_addr.key_nr = dns_rr_dst.dns_addr_key_nr
                         WHERE dns_addr.dns_range_key_nr = {range_nr} and dns_addr.is_rr_usable and dns_rr_dst.key_nr is NULL and dns_addr.ip_addr::inet > {curr_addr}::text::inet ORDER BY dns_addr.ip_addr LIMIT 1
                         """, {'range_nr': self.key_nr, 'curr_addr': str(addr)})
        if len(res) == 1:
            return res[0]['ip_addr']
        else:
            return None

    def get_n_trees(self, db, connection):
        res = db.execute(connection,
                         "SELECT dns_ntree.* FROM dns_ntree INNER JOIN dns_domain2range ON dns_ntree.key_nr = dns_domain2range.dns_ntree_key_nr "
                         "WHERE dns_domain2range.dns_range_key_nr = {dns_range} ORDER by domain_order ASC",
                         {'dns_range': self.key_nr})
        return [NTree(**r) for r in res]

    def get_rr(self, db, connection, login_name):
        res = db.execute(connection,
                         """
                         SELECT
                            (rrt_rec).name as META_rr_rr_type,
                            (rrt_rec).force_auto_dml as META_rr_force_auto_dml,
                            host_is_nws as META_rr_host_is_nws,
                            fqdn as dns_ntree_fqdn,
                            fqdn_description as dns_ntree_description,
                            rr_data as META_rr_rr_data,
                            rr_data_unref as META_rr_rr_data_unref,
                            rr_data_target_ipaddr as META_rr_rr_data_target_ipaddr,
                            rr_data_target_fqdn as META_rr_rr_data_target_fqdn,
                            (dbrt_rec).name as META_rr_dbrt_name,
                            (fqdn_dbnt_rec).is_d as dns_ntree_is_d,
                            (fqdn_dbnt_rec).name as dns_ntree_dbnt_name
                         FROM dns.list_rr(in_login_name:={login_name},
                            in_range_list:={range},
                            in_force_range_owner:=TRUE,
                            in_listopt_global:={global})
                         ORDER BY (rrt_rec).name, dns_ntree_fqdn, rr_data_target_ipaddr, META_rr_rr_data
                         """,
                         {'range': [self.print_name], 'login_name': login_name,
                          'global': DBMgr.get_by_login_name(login_name=login_name, connection=connection,
                                                            db=db).has_permission('dns.ro')})  # FIXME?
        return [{'NTree': NTree(**_marshal_dict(r, NTree)), 'rr': RR(**_marshal_dict(r, RR))} for r in
                res]

    def get_nat_range(self, db, connection):
        if self._nat_range is not None:
            return self._nat_range
        res = db.execute(connection,
                         """
                            SELECT dns_range_l_key_nr
                            FROM
                            nat_range2range
                            WHERE
                            dns_range_r_key_nr = {key_nr}
                         """, {'key_nr': self.key_nr})
        if len(res) == 0:
            return None
        l_id = res[0]['dns_range_l_key_nr']
        if l_id == self.key_nr:
            self._nat_range = self
            return self
        self._nat_range = DBNetArea.get_by_key_nr(db=db, connection=connection, key_nr=l_id)
        return self._nat_range

    def get_gateways(self, db, connection):
        return [r['list_gateways'] for r in db.execute(connection, "SELECT * from dns.list_gateways({range})",
                                                       {'range': self.print_name})]

    def get_addrs(self, db, connection):
        res = db.execute(connection, "SELECT * FROM dns_addr WHERE dns_range_key_nr = {range}", {'range': self.key_nr})
        return [DNSAddr(**r) for r in res]

    def get_max_addr(self, db, connection):
        res = db.execute(connection,
                         "SELECT * FROM dns_addr WHERE dns_range_key_nr = {range} ORDER by ip_addr DESC LIMIT 1",
                         {'range': self.key_nr})
        return DNSAddr(**res[0]) if len(res) > 0 else None

    def get_min_addr(self, db, connection):
        res = db.execute(connection,
                         "SELECT * FROM dns_addr WHERE dns_range_key_nr = {range} ORDER by ip_addr ASC LIMIT 1",
                         {'range': self.key_nr})
        return DNSAddr(**res[0]) if len(res) > 0 else None

    def get_reserved_addrs(self, db, connection):
        res = db.execute(connection,
                         "SELECT * FROM dns_addr WHERE dns_range_key_nr = {range} AND is_reserved order by ip_addr",
                         {'range': self.key_nr})
        return [DNSAddr(**r) for r in res]

    @staticmethod
    def get_area_utilization(db, connection, login, areas):
        rest = dict()
        a_by_id = {a.key_nr: a for a in areas}
        res_reserved = db.execute(connection,
                                  """
                                  select dns_range_key_nr as k, count(DISTINCT ip_addr) as c -- distinct: mehrdeutige a-rr 
                                  from dns_addr 
                                      left join dns_rr_dst on dns_addr.key_nr=dns_rr_dst.dns_addr_key_nr 
                                      left join dns_rr_type on dns_rr_dst.dns_rr_type_key_nr=dns_rr_type.key_nr 
                                  WHERE (dns_rr_type.name='A' or dns_rr_type.name='AAAA' or dns_addr.is_reserved) 
                                      and dns_addr.dns_range_key_nr = ANY({range}) 
                                  GROUP BY dns_range_key_nr

                                """,
                                  {'range': [int(a.key_nr) for a in areas]})
        for r in res_reserved:
            area = a_by_id[r['k']]
            area = area.print_name
            rest[area] = rest.get(area, 0) + r['c']
        return rest  # FIXME: Propably wrong


class Facility(DBObject):
    TABLE = "vw_oe"

    def __init__(self, key_nr=None, parent_key_nr=None, vw_oe_typ_key_nr=None, name=None, short_name=None,
                 sub_sortby=None, **args):
        self.key_nr = key_nr
        self.name = name
        self.short_name = short_name


class Net(DBObject):
    TABLE = "nd_ip_subnet"

    def __init__(self, net=None, key_nr=None, nd_net_instnc_is_priv=None, nd_bcd_key_nr=None, subnet_is_stat_bound=None,
                 subnet_type=None, descr=None, **args):
        self.key_nr = key_nr
        self.net = net
        self.descr = descr


class RR(MetaDBObject):
    TABLE = 'meta_rr'

    def __init__(self, rr_data, rr_data_unref, dbrt_name, rr_data_target_ipaddr=None, rr_data_target_fqdn=None,
                 rr_type=None, host_is_nws=None, force_auto_dml=False, **args):
        self.rr_data = rr_data
        self.force_auto_dml = force_auto_dml
        self.rr_data_unref = rr_data_unref
        self.rr_data_target = rr_data_target_ipaddr if rr_data_target_fqdn is None else dns.name.from_text(
            rr_data_target_fqdn).to_unicode()
        self.host_is_nws = host_is_nws
        self.inttype = IntType(name=dbrt_name, description=None, dns_rr_type_name=rr_type)

    @staticmethod
    def get(fqdn, inttype, rr_data, login_name, db, connection):
        res = db.execute(connection,
                         """
                         SELECT 
                            (rrt_rec).name as META_rr_rr_type, 
                            (rrt_rec).force_auto_dml as META_rr_force_auto_dml,
                            host_is_nws as META_rr_host_is_nws, 
                            fqdn as dns_ntree_fqdn, 
                            fqdn_description as dns_ntree_description, 
                            rr_data as META_rr_rr_data,
                            rr_data_unref as META_rr_rr_data_unref,
                            rr_data_target_ipaddr as META_rr_rr_data_target_ipaddr,
                            rr_data_target_fqdn as META_rr_rr_data_target_fqdn, 
                            (dbrt_rec).name as META_rr_dbrt_name
                         FROM dns.list_rr(in_login_name:={login_name},
                            in_listopt_global:={global}, in_fqdn:={fqdn}, in_inttype_list:={inttype}) WHERE
                            rr_data={rr_data}
                         ORDER BY rr_data_target_ipaddr, META_rr_rr_data""",
                         {'global': DBMgr.get_by_login_name(login_name=login_name, connection=connection,
                                                            db=db).has_permission('dns.ro'),
                          'login_name': login_name, 'fqdn': fqdn, 'inttype': [inttype],
                          'rr_data': rr_data})
        if len(res) == 0:
            return None
        return {'NTree': NTree(**_marshal_dict(res[0], NTree)), 'rr': RR(**_marshal_dict(res[0], RR))}

    @staticmethod
    def get_by_target_fqdn(login_name, target_fqdn, db, connection):
        res = db.execute(connection,
                         "SELECT fqdn as dns_ntree_fqdn, (rrt_rec).name as META_rr_rr_type, rr_data as META_rr_rr_data,"
                         "(rrt_rec).force_auto_dml as META_rr_force_auto_dml,"
                         "rr_data_target_ipaddr as META_rr_rr_data_target_ipaddr,"
                         "(dbrt_rec).name as META_rr_dbrt_name,"
                         "rr_data_unref as META_rr_rr_data_unref,"
                         "(select array_agg(name || suffix) from dns_range where key_nr = any(target_range_pk_list)) as META_range_list,"
                         "host_is_nws as META_rr_host_is_nws, "
                         "fqdn_description as dns_ntree_description,"
                         "rr_data_target_fqdn as META_rr_rr_data_target_fqdn, "
                         "(fqdn_dbnt_rec).name as dns_ntree_dbnt_name FROM "
                         "dns.list_rr(in_login_name:={login_name}, in_target_fqdn:={target_fqdn}, in_listopt_global:={global}) "
                         "ORDER BY META_rr_rr_type",
                         {'login_name': login_name, 'target_fqdn': target_fqdn,
                          'global': DBMgr.get_by_login_name(login_name=login_name, connection=connection,
                                                            db=db).has_permission('dns.ro')})
        return [{
            'nws_range': (
                DBNetArea.get_by_name(print_name=r['meta_range_list'][0], db=db, connection=connection).get_nat_range(
                    db=db,
                    connection=connection)
                if r['meta_range_list'] is not None and r['meta_rr_host_is_nws'] else None
            ),
            'rr': RR(**_marshal_dict(r, RR)), 'NTree': NTree(**_marshal_dict(r, NTree)),
            'ranges': (r['meta_range_list'] if r['meta_range_list'] is not None else list())} for r in res]

    @staticmethod
    def get_by_fqdn(login_name, fqdn, db, connection):
        res = db.execute(connection,
                         "SELECT fqdn as dns_ntree_fqdn, (rrt_rec).name as META_rr_rr_type,"
                         "(rrt_rec).force_auto_dml as META_rr_force_auto_dml,"
                         "rr_data as META_rr_rr_data,"
                         "rr_data_target_ipaddr as META_rr_rr_data_target_ipaddr,"
                         "(dbrt_rec).name as META_rr_dbrt_name,"
                         "host_is_nws as META_rr_host_is_nws, "
                         "rr_data_unref as META_rr_rr_data_unref,"
                         "(select array_agg(name || suffix) from dns_range where key_nr = any(target_range_pk_list)) as META_range_list,"
                         "fqdn_description as dns_ntree_description,"
                         "rr_data_target_fqdn as META_rr_rr_data_target_fqdn, "
                         "(fqdn_dbnt_rec).name as dns_ntree_dbnt_name FROM "
                         "dns.list_rr(in_login_name:={login_name}, in_fqdn:={fqdn}, in_listopt_global:={global}) "
                         "ORDER BY META_rr_rr_type",
                         {'login_name': login_name, 'fqdn': fqdn,
                          'global': DBMgr.get_by_login_name(login_name=login_name, connection=connection,
                                                            db=db).has_permission('dns.ro')})
        return [{
            'nws_range': (
                DBNetArea.get_by_name(print_name=r['meta_range_list'][0], db=db, connection=connection).get_nat_range(
                    db=db,
                    connection=connection)
                if r['meta_range_list'] is not None and r['meta_rr_host_is_nws'] else None
            ),
            'rr': RR(**_marshal_dict(r, RR)), 'NTree': NTree(**_marshal_dict(r, NTree)),
            'ranges': (r['meta_range_list'] if r['meta_range_list'] is not None else list())} for r in res]


class NTree(DBObject):
    TABLE = "dns_ntree"

    def __init__(self, key_nr=None, parent_key_nr=None, parent_fqdn=None, parent_is_d=None, is_d=None, is_a=None,
                 is_dhcp=None, dbnt_name=None,
                 is_rad=None, is_wildcard=None, dns_name_type_name=None, name=None, description=None, fqdn=None,
                 fqdn_hash=None, parent_zone_fqdn_hash=None, dns_zone_key_nr=None, dns_parent_zone_key_nr=None,
                 hierarchy_pk_list=None, fqdn_sub_count=None, fqdn_has_rr=None, **args):
        self.key_nr = key_nr
        self.parent_key_nr = parent_key_nr
        self.parent_is_d = parent_is_d
        self.is_d = is_d
        self.is_a = is_a
        self.inttype = dbnt_name
        self.is_dhcp = is_dhcp
        self.is_wildcard = is_wildcard
        self.name = name
        self.description = description if not description == '' else None
        self.fqdn = dns.name.from_text(fqdn).to_unicode() if fqdn else None
        self.dns_zone_key_nr = dns_zone_key_nr
        self.dns_parent_zone_key_nr = dns_parent_zone_key_nr
        self.parent_fqdn = dns.name.from_text(parent_fqdn).to_unicode() if parent_fqdn and not fqdn == '.' else None
        self.fqdn_sub_count = fqdn_sub_count
        self.fqdn_has_rr = fqdn_has_rr
        self.dbnt_name = dbnt_name

    def update_fqdn_description(self, descr):
        if self.description == descr:
            return None
        return self.update_fqdn_description_deferrable(old_description=self.description, new_description=descr,
                                                       fqdn=self.fqdn)

    def update(self, new_fqdn, new_description, new_inttype):
        return NTree.update_deferrable(old_fqdn=self.fqdn, new_fqdn=new_fqdn, old_description=self.description,
                                       new_description=new_description, old_inttype=self.inttype,
                                       new_inttype=new_inttype)


class DNSAddr(DBObject):
    TABLE = 'dns_addr'

    def __init__(self, key_nr=None, ip_addr=None, addr_type=None, is_reserved=None, is_rr_usable=None, is_dhcp=None,
                 dns_range_key_nr=None, nd_ip_subnet_key_nr=None, **args):
        self.key_nr = key_nr
        self.ip_addr = ip_addr
        self.dns_range_ley_nr = dns_range_key_nr
        self.is_reserved = is_reserved
        self.is_dhcp = is_dhcp
        self.is_rr_usable = is_rr_usable

    def __eq__(self, other):
        return (type(other) == str and other == str(self.ip_addr)) or (
                type(self) == type(other) and self.key_nr == other.key_nr)


class NDPPort(DBObject):
    TABLE = 'nd_p_port'

    def __init__(self, name, module=None, **kwargs):
        self.name = name
        self.module = module

    @staticmethod
    def get_nd_p_port_by_vlan(db, connection, vlan_id, ni_name):
        res = db.execute(connection,
                         """
                         select nd_mdl_typ_class.name as root_nd_mdl_typ_class_name, root_typ.name as root_nd_mdl_typ_name, dst.name as nd_p_port_name, root.name as root_nd_mdl_name, mod.name as base_nd_mdl_name,
nd_room.name as nd_room_name, nd_room.nr as nd_room_nr, nd_floor.nr as nd_floor_nr, nd_floor.descr as nd_floor_descr,
nd_bldg.nr as nd_bldg_nr, nd_bldg.name as nd_bldg_name from nd_dev
  INNER JOIN dns_ntree on nd_dev.dns_ntree_key_nr=dns_ntree.key_nr
  inner join nd_l_port on nd_dev.key_nr=nd_l_port.nd_dev_key_nr
  inner join nd_vlan on nd_l_port.nd_vlan_key_nr=nd_vlan.key_nr
  inner join nd_net_instnc on nd_vlan.nd_net_instnc_key_nr=nd_net_instnc.key_nr
  inner join nd_l2p_port on nd_l_port.key_nr=nd_l2p_port.nd_l_port_key_nr
  inner join nd_p_port src on nd_l2p_port.nd_p_port_key_nr = src.key_nr
  inner join nd_p_port dst on src.nd_dest_p_port_key_nr=dst.key_nr
  inner join nd_mdl mod on dst.nd_mdl_key_nr=mod.key_nr
  inner join nd_mdl root on mod.nd_root_mdl_key_nr=root.key_nr
  inner join nd_room on root.nd_room_key_nr=nd_room.key_nr
  inner join nd_floor on nd_room.nd_floor_key_nr=nd_floor.key_nr
  inner join nd_bldg on nd_room.nd_bldg_key_nr=nd_bldg.key_nr
  inner join nd_mdl_typ root_typ on root.nd_mdl_typ_key_nr = root_typ.key_nr
  inner join nd_mdl_typ_class on root_typ.nd_mdl_typ_class_key_nr=nd_mdl_typ_class.key_nr
  where nd_net_instnc.name = {ni_name} and nd_vlan.id = {vlan_id}
                         """,
                         {'ni_name': ni_name, 'vlan_id': vlan_id})
        if len(res) == 0:
            return None
        return [NDPPort(
            **_marshal_dict(r, NDPPort),
            module=NDModule(**_marshal_dict(_marshal_dict_str(r, 'base'), NDModule),
                            root_module=NDModule(**_marshal_dict(_marshal_dict_str(r, 'root'), NDModule),
                                                 room=NDRoom(
                                                     **_marshal_dict(r, NDRoom)),
                                                 floor=NDFloor(
                                                     **_marshal_dict(r, NDFloor)),
                                                 building=NDBuilding(
                                                     **_marshal_dict(r, NDBuilding)),
                                                 module_type=NDModuleType(
                                                     **_marshal_dict(_marshal_dict_str(r, 'root'), NDModuleType),
                                                     type_class=NDModuleTypeClass(
                                                         **_marshal_dict(_marshal_dict_str(r, 'root'),
                                                                         NDModuleTypeClass)
                                                     )
                                                 )
                                                 ),
                            )
        ) for r in res]

    @staticmethod
    def get_nd_p_port_by_dev_fqdn_and_port_name(db, connection, fqdn, port):
        res = db.execute(connection,
                         """
select nd_mdl_typ_class.name as root_nd_mdl_typ_class_name, root_typ.name as root_nd_mdl_typ_name, dst.name as nd_p_port_name, root.name as root_nd_mdl_name, mod.name as base_nd_mdl_name,
nd_room.name as nd_room_name, nd_room.nr as nd_room_nr, nd_floor.nr as nd_floor_nr, nd_floor.descr as nd_floor_descr, 
nd_bldg.nr as nd_bldg_nr, nd_bldg.name as nd_bldg_name from nd_dev
  INNER JOIN dns_ntree on nd_dev.dns_ntree_key_nr=dns_ntree.key_nr
  inner join nd_l_port on nd_dev.key_nr=nd_l_port.nd_dev_key_nr
  inner join nd_l2p_port on nd_l_port.key_nr=nd_l2p_port.nd_l_port_key_nr
  inner join nd_p_port src on nd_l2p_port.nd_p_port_key_nr = src.key_nr
  inner join nd_p_port dst on src.nd_dest_p_port_key_nr=dst.key_nr
  inner join nd_mdl mod on dst.nd_mdl_key_nr=mod.key_nr
  inner join nd_mdl root on mod.nd_root_mdl_key_nr=root.key_nr
  inner join nd_room on root.nd_room_key_nr=nd_room.key_nr
  inner join nd_floor on nd_room.nd_floor_key_nr=nd_floor.key_nr
  inner join nd_bldg on nd_room.nd_bldg_key_nr=nd_bldg.key_nr
  inner join nd_mdl_typ root_typ on root.nd_mdl_typ_key_nr = root_typ.key_nr
  inner join nd_mdl_typ_class on root_typ.nd_mdl_typ_class_key_nr=nd_mdl_typ_class.key_nr
where dns_ntree.fqdn = {fqdn} and nd_l_port.name={port}
                         """,
                         {'port': port, 'fqdn': fqdn})
        if len(res) == 0:
            return None
        res = res[0]
        return NDPPort(
            **_marshal_dict(res, NDPPort),
            module=NDModule(**_marshal_dict(_marshal_dict_str(res, 'base'), NDModule),
                            root_module=NDModule(**_marshal_dict(_marshal_dict_str(res, 'root'), NDModule),
                                                 room=NDRoom(
                                                     **_marshal_dict(res, NDRoom)),
                                                 floor=NDFloor(
                                                     **_marshal_dict(res, NDFloor)),
                                                 building=NDBuilding(
                                                     **_marshal_dict(res, NDBuilding)),
                                                 module_type=NDModuleType(
                                                     **_marshal_dict(_marshal_dict_str(res, 'root'), NDModuleType),
                                                     type_class=NDModuleTypeClass(
                                                         **_marshal_dict(_marshal_dict_str(res, 'root'),
                                                                         NDModuleTypeClass)
                                                     )
                                                 )
                                                 ),
                            )
        )


class NDBuilding(DBObject):
    TABLE = 'nd_bldg'

    def __init__(self, nr, name=None, **kwargs):
        self.nr = nr
        self.name = name


class NDFloor(DBObject):
    TABLE = 'nd_floor'

    def __init__(self, nr, descr=None, **kwargs):
        self.nr = nr
        self.description = descr


class NDRoom(DBObject):
    TABLE = 'nd_room'

    def __init__(self, nr, name=None, **kwargs):
        self.nr = nr
        self.name = name


class NDModule(DBObject):
    TABLE = 'nd_mdl'

    def __init__(self, name, descr=None, root_module=None, room=None, floor=None, building=None, module_type=None,
                 **kwargs):
        self.name = name
        self.description = descr
        self.root_module = root_module
        self.room = room
        self.building = building
        self.floor = floor
        self.module_type = module_type


class NDModuleType(DBObject):
    TABLE = 'nd_mdl_typ'

    def __init__(self, name, type_class=None, **kwargs):
        self.name = name
        self.type_class = type_class


class NDModuleTypeClass(DBObject):
    TABLE = 'nd_mdl_typ_class'

    def __init__(self, name, **kwargs):
        self.name = name


class APIToken(MetaDBObject):

    def __init__(self, pk=None, token=None, login_name=None, description=None):
        self.token = token
        self.login_name = login_name
        self.description = description
        self.pk = pk
        if id is None and token is not None:
            self.id = int(token.split('.')[0])

    def delete(self, db, conn):
        query = """
               select wapi_3_0.ta_handler(
                in_login_name => {{login_name}},
                in_stmt_list => array[row(
                 'cntl',
                 'wapi_auth',
                 'delete',
                 array['{dict}']::json[],
                 '{{}}'
                 )::wapi_3_0.imp_ta_stmt_rec_type],
                in_report_stmt_pos => false,
                in_is_dry_mode => false
                );
        """.format(dict=json.dumps({'name': 'pk', 'old_value': self.pk}).replace('{', '{{').replace('}', '}}')).replace(
            '{}', '{{}}')
        print(query)
        db.execute(conn,
                   query, {'login_name': self.login_name})
        self.token = None
        self.pk = None
        self.description = None
        self.login_name = None


# Pseudo-objects for strings

class VDNSRecords(DBObject):
    TABLE = 'pdns_rr'


# API objects

class APINetArea(NetArea):
    def __init__(self, name, suffix, facility_short, net):
        self.name = name
        self.suffix = suffix
        self.facility = Facility(short_name=facility_short)
        self.net = Net(net=ipaddress.IPv6Network(net) if ':' in net else ipaddress.IPv4Network(net))


class APIMgr(Mgr):
    __areas__ = list()

    @staticmethod
    def get_by_login_name(login_name, api_cert, api_key):
        r = requests.get("https://www-net.scc.kit.edu/api/2.0/dns/mgr/list", params={'login_name': login_name},
                         cert=(api_cert, api_key))
        res = r.json()
        if len(res) == 0:
            return None
        mgr = APIMgr(**res[0]['mgr'])
        mgr.__areas__ = [
            APINetArea(name=r['range'].split('/')[0] + '/', suffix=int(r['range'].split('/')[1]),
                       net=r['subnet_cidr_spec'],
                       facility_short=r['ou']) for r in res]
        return mgr

    def get_areas(self, *args, **kwargs):
        return self.__areas__
