from flask import request, redirect, url_for, render_template, g, abort, flash, jsonify
from functools import wraps
import postgresql.exceptions
import json
from . import app, db, VERSION, MODS


def get_db_conn():
    if not hasattr(g, 'kit_pg_db_conn'):
        g.kit_pg_db_conn = db.connect()
        # g.ta = g.kit_pg_db_conn.xact(mode='read only')
        g.ta = g.kit_pg_db_conn.xact()
        if 'login' in request.environ['beaker.session']:
            s = request.environ['beaker.session']['login']
            s.update_acd_cache(db=db, connection=get_db_conn())
            request.environ['beaker.session'].save()
        g.ta.begin()
    return g.kit_pg_db_conn


@app.after_request
def apply_csp(response):
    csp = "default-src 'self'; script-src 'self'"
    response.headers['Content-Security-Policy'] = csp
    response.headers['X-Content-Security-Policy'] = csp
    response.headers['X-WebKit-CSP'] = csp
    response.headers['X-Frame-Options'] = "SAMEORIGIN"
    response.headers['X-Xss-Protection'] = "1; mode=block"
    response.headers['X-Content-Type-Options'] = "nosniff"
    response.headers['Referrer-Policy'] = "same-origin"
    return response


@app.teardown_appcontext
def close_db(error):
    if hasattr(g, 'kit_pg_db_conn'):
        try:
            if not g.ta.state == 'initialized':
                g.ta.commit()
        except postgresql.exceptions.Exception as e:
            print('got error on close... ignoring', e)
        g.kit_pg_db_conn.close()


def api_login_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        if 'login' not in request.environ['beaker.session']:
            return jsonify({'error': 'Authentication required'}), 403
        return f(*args, **kwargs)

    return decorated_function


def login_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        if 'login' not in request.environ['beaker.session']:
            return redirect(url_for(app.config.get('LOGIN_URL'), next=request.url))
        return f(*args, **kwargs)

    return decorated_function


@app.route('/')
def main():
    return redirect(url_for(app.config.get('DEFAULT_URL')))


@app.route('/api/system/info')
def api_sysinfo():
    return jsonify({'version': VERSION, 'mods': MODS, 'host_oper_mode': db.host_omdl, 'js_mods': app.config.get('JS_MODULES')})


@app.route('/api/session')
def api_session():
    s = request.environ['beaker.session']
    if 'login' in s:
        return jsonify({'user': s['login']})
    return jsonify({'error': 'No session'}), 403


@app.route('/logout', methods=['GET', 'POST'])
def logout():
    s = request.environ['beaker.session']
    if 'login' in s:
        s.delete()
    if 'dst' in request.args:
        return redirect(url_for(app.config.get('LOGIN_URL'), dst=request.args['dst']))
    return redirect('/')


@app.errorhandler(postgresql.exceptions.Exception)
def handle_db_error(e):
    try:  # TODO: val_wi &co in tabelle aufbereiten
        err = json.loads(e.details['hint'])
        return jsonify({'type': 'db_error', 'error': err}), 500
    except:
        if hasattr(e, 'details'):
            return jsonify({'type': 'generic_error', 'error': e.details}), 500
        return jsonify({'type': 'generic_error', 'error': e}), 500


@app.errorhandler(Exception)
def handle_generic(e):
    if app.config.get('DEBUG', False):
        raise e
    if hasattr(e, 'details'):
        return jsonify({'type': 'generic_error', 'error': e.details}), 500
    return jsonify({'type': 'generic_error', 'error': str(e)}), 500


